package dictionarybuilder;

import java.io.*;
import java.util.*;

/**
 * Предоставляет доступ к множеству маленьких файлов, находящихся в одной
 * директории точно к одному большому файлу. Максимальный размер одного члена
 * колонии задан в {@link EasyAccessFilesColony#MAX_FILE_SIZE}. По мере
 * чтения/записи в колонию файлов, открываются нужные файлы-члены колонии. При
 * выполнении {@link EasyAccessFilesColony#seek(long)}, а затем записи за
 * пределы колонии/одного её члена, создаётся необходимый файл, а пропущенное
 * пространство в пределах одного файла оказывается полным случайных данных.
 * Членами колонии являются файлы с именами-шестнадцатиричными числами (первым
 * будет файл с именем "0").
 *
 * @author quasiyoke
 */
public class EasyAccessFilesColony extends EasyAccessFile{

private static final int MAX_FILE_SIZE=20*1024;

private Vector<RandomAccessFile> files=new Vector<RandomAccessFile>();

private String folderPath;

/**
 * Если <code>folder</code> -- не папка, файл удаляется и создаётся папка с
 * таким же именем.
 */
public EasyAccessFilesColony(File folder,boolean writingResolved) throws
	IOException{
    super(folder,writingResolved);
    if(!folder.isDirectory())
	if(writingResolved){
	    if(folder.isFile())
		folder.delete();
	    folder.mkdirs();
	}else
	    throw new FileNotFoundException();
    folderPath=folder.getPath();
}

public void close() throws IOException{
    int j;
    int filesCount=files.size();
    for(j=0;j<filesCount;j++)
	files.get(j).close();
}

private RandomAccessFile getDataInput() throws IOException{
    RandomAccessFile file;
    int fileNumber=(int)(filePointer/MAX_FILE_SIZE);
    boolean resizingNeeds=fileNumber>=files.size();
    if(resizingNeeds||(file=files.get(fileNumber))==null){
	if(resizingNeeds)
	    files.setSize(fileNumber+1);
	try{
	    files.set(fileNumber,file=new RandomAccessFile(folderPath+
		    File.separatorChar+Integer.toHexString(fileNumber),writingResolved
		    ?"rw":"r"));
	}catch(FileNotFoundException e){
	    throw new EOFException();
	}
    }
    file.seek(filePointer%MAX_FILE_SIZE);
    return file;
}

private RandomAccessFile getDataOutput() throws IOException{
    if(writingResolved){
	RandomAccessFile file;
	int fileNumber=(int)(filePointer/MAX_FILE_SIZE);
	boolean resizingNeeds=fileNumber>=files.size();
	if(resizingNeeds||(file=files.get(fileNumber))==null){
	    if(resizingNeeds)
		files.setSize(fileNumber+1);
	    files.set(fileNumber,file=new RandomAccessFile(folderPath+
		    File.separatorChar+Integer.toHexString(fileNumber),"rw"));
	}
	file.seek(filePointer%MAX_FILE_SIZE);
	return file;
    }
    return null;
}

public char readChar() throws IOException{
    return (char)((readUnsignedByte()<<8)|readUnsignedByte());
}

public int readInt() throws IOException{
    return (readUnsignedByte()<<24)|(readUnsignedByte()<<16)|
	    (readUnsignedByte()<<8)|readUnsignedByte();
}

public int readUnsignedByte() throws IOException{
    int result=getDataInput().readUnsignedByte();
    filePointer++;
    return result;
}

public void seek(long position) throws IOException{
    filePointer=position;
}

public void skipBytes(int bytesCount) throws IOException{
    if(bytesCount>0)
	filePointer+=bytesCount;
}

public void writeByte(int nByte) throws IOException{
    getDataOutput().writeByte(nByte);
    filePointer++;
}

public void writeChar(char character) throws IOException{
    writeByte(character>>8);
    writeByte(character);
}

public void writeInt(int nByte) throws IOException{
    writeByte(nByte>>24);
    writeByte(nByte>>16);
    writeByte(nByte>>8);
    writeByte(nByte);
}

}
