package dictionarybuilder;

import java.io.*;
import java.util.*;

/**
 * Преобразователь словарей от MobileLexicon
 * (volgosoft.narod.ru/mdictionary.htm) в формат, приемлемый для FreshDictionary.
 * Для обработки нужно сообщить язык ключей и статей словаря с помощью ISO 639-2
 * (http://www.sil.org/iso639-3/codes.asp?order=639_2&letter=%25), а также
 * информацию о том, где был найден материал для этого словаря в файле
 * SOURCE_FOLDER_PATH\information.txt через перенос строки в кодировке UTF-8.
 *
 * @author quasiyoke
 */
public class Main{

/**
 * Разделитель ключа к статье и её текста.
 */
private final static char ARTICLE_KEY_AND_ARTICLE_TEXT_DIVIDER=1;

public final static int BYTE_LENGTH=1;

public final static int CHAR_LENGTH=Character.SIZE/Byte.SIZE;

private final static char DESTINATION_FILE_ARTICLE_ENDING=0;

public final static int INT_LENGTH=Integer.SIZE/Byte.SIZE;

public final static int NULL_POINTER=0xFFFFFFFF;

/**
 * Символ, которым кончаются статьи в файлах исходного словаря.
 */
private final static char SOURCE_FILE_ARTICLE_ENDING='*';

private static File articlesFile;

private static EasyAccessSolitaryFile articlesFileAccess;


static{
    try{
	articlesFile=File.createTempFile("articles",null);
	articlesFile.deleteOnExit();
	articlesFileAccess=new EasyAccessSolitaryFile(articlesFile,true);
    }catch(IOException e){
	System.err.println("Error in static{} block.");
    }
}

/**
 * @param args 
 *            Аргументы командной строки.
 */
public static void main(String[] args) throws IOException{
    int j;
    int i;
    String dictionaryKeysLanguage;
    String dictionaryArticlesLanguage;
    String dictionarySource;
    if(args.length==3&&args[0].equalsIgnoreCase("-f")){
	File destinationFolder=new File(args[2]);
	EasyAccessFilesColony destinationFileAccess;
	try{
	    destinationFileAccess=new EasyAccessFilesColony(destinationFolder,
		    false);
	}catch(FileNotFoundException e){
	    System.err.println("Error, file colony (folder) \""+
		    destinationFolder.getPath()+
		    "\" should exist.");
	    return;
	}
	String word=args[1].trim().toLowerCase();
	dictionaryKeysLanguage=destinationFileAccess.readUTF8String();
	dictionaryArticlesLanguage=destinationFileAccess.readUTF8String();
	dictionarySource=destinationFileAccess.readUTF8String();
	int alphabetSize=destinationFileAccess.readUnsignedByte();
	char[] alphabet=new char[alphabetSize];
	for(j=0;j<alphabetSize;j++)
	    alphabet[j]=destinationFileAccess.readChar();
	int currentSubindexSubindexesCount=destinationFileAccess.
		readUnsignedByte();
	int subindexSubindexesCount=0;
	boolean subindexFound;
	int offset;
	int wordLength=word.length();
	boolean freshFreeSubindexNeeds;
	int characterIndex;
	int subindexOffset=(int)destinationFileAccess.getFilePointer();
	char wordCharacter;
	int someSubindexSubindexesCount;
	Vector<Integer> freeSubindexOffset=new Vector<Integer>();
	for(j=0;j<wordLength;j++){
	    wordCharacter=word.charAt(j);
	    subindexFound=false;
	    destinationFileAccess.seek(subindexOffset+INT_LENGTH*2);
	    offset=destinationFileAccess.readInt();
	    if(offset==NULL_POINTER){
		if(printFreeArticles(destinationFileAccess,freeSubindexOffset)==
			0)
		    System.out.println("No articles found.");
		return;
	    }
	    freshFreeSubindexNeeds=(offset&Index.HAVE_NOT_FREE_SUBINDEX_MASK)==0;
	    offset&=Index.SUBINDEXES_OFFSET_MASK;
	    for(i=0;i<currentSubindexSubindexesCount&&(!subindexFound||
		    freshFreeSubindexNeeds);i++){
		characterIndex=destinationFileAccess.readUnsignedByte();
		someSubindexSubindexesCount=destinationFileAccess.
			readUnsignedByte();
		if(characterIndex==0xFF){
		    freeSubindexOffset.add(offset);
		    freshFreeSubindexNeeds=false;
		}else if((!subindexFound)&&(subindexFound=
			(alphabet[characterIndex]==wordCharacter))){
		    subindexOffset=offset;
		    subindexSubindexesCount=someSubindexSubindexesCount;
		}
		offset+=Index.length(someSubindexSubindexesCount);
	    }
	    if(freshFreeSubindexNeeds)
		System.err.println("Error, no free subindexes found (\""+word+
			"\", offset "+j+").");
	    if(subindexFound)
		currentSubindexSubindexesCount=subindexSubindexesCount;
	    else{
		if(printFreeArticles(destinationFileAccess,
			freeSubindexOffset)==0)
		    System.out.println("No articles found.");
		return;
	    }
	}
	destinationFileAccess.seek(subindexOffset+INT_LENGTH);
	if((offset=destinationFileAccess.readInt())==NULL_POINTER){
	    if(printFreeArticles(destinationFileAccess,
		    freeSubindexOffset)==0)
		System.out.println("No articles found.");
	    return;
	}
	printArticle(destinationFileAccess,offset);
	printFreeArticles(destinationFileAccess,freeSubindexOffset);
	return;
    }
    if(args.length!=2){
	printHelp();
	return;
    }
    File destinationFolder=new File(args[1]);
    File sourceFolder=new File(args[0]);
    if(!sourceFolder.exists()||sourceFolder.isFile()){
	printHelp();
	return;
    }
    removeAnyFile(destinationFolder);
    if(destinationFolder.isDirectory())
	if(destinationFolder.exists())
	    if(destinationFolder.isDirectory()){
		printHelp();
		return;
	    }else{
		System.out.print("Overwrite \""+destinationFolder.getPath()+
			"\"? (y/n): ");
		if(Character.toLowerCase(System.in.read())!='y')
		    return;
		destinationFolder.delete();
	    }
	else{
	    String destinationFileParentPath;
	    if((destinationFileParentPath=destinationFolder.getParent())!=null)
		(new File(destinationFileParentPath)).mkdirs();
	}
    String sourceIndexPath=
	    sourceFolder.getPath()+File.separatorChar+"letter.txt";
    File sourceIndex=new File(sourceIndexPath);
    EasyAccessSolitaryFile sourceIndexAccess;
    try{
	sourceIndexAccess=new EasyAccessSolitaryFile(sourceIndex,false);
    }catch(FileNotFoundException e){
	System.err.println("Error, file \""+sourceIndexPath+"\" should exist.");
	return;
    }
    int nCharacter;
    char character;
    StringBuffer sbLetterIndex;
    String letterFolderPath;
    String sLetterIndex;
    File letterFolder;
    String wordsFolderPath;
    String wordsTranslationsFolderPath;
    File wordsFolder;
    while(true)
	try{
	    /*
	     * Читаем букву и тут же пропускаем двоеточие.
	     */
	    character=sourceIndexAccess.readUTF8Char();
	    sourceIndexAccess.readUnsignedByte();
	    sbLetterIndex=new StringBuffer();
	    while((nCharacter=sourceIndexAccess.readUnsignedByte())!=
		    0x0D){
		if(nCharacter==-1)
		    break;
		if('0'<=nCharacter&&nCharacter<='9')
		    sbLetterIndex.append((char)nCharacter);
	    }
	    sLetterIndex=sbLetterIndex.toString();
	    try{
		Index.addToAlphabet(Integer.parseInt(sLetterIndex),
			character);
	    }catch(NumberFormatException e){
		System.err.println("Error, \""+sLetterIndex+
			"\" isn't like a number at \""+sourceIndexPath+"\".");
		continue;
	    }
	    /*
	     * Пропускаем 0x0A -- возврат каретки из
	     * последовательности "перенос строки, возврат каретки".
	     */
	    sourceIndexAccess.readUnsignedByte();
	}catch(EOFException e){
	    break;
	}
    EasyAccessFilesColony destinationFileAccess=new EasyAccessFilesColony(
	    destinationFolder,true);
    String informationFilePath=sourceFolder.getPath()+File.separatorChar+
	    "information.txt";
    try{
	EasyAccessSolitaryFile informationFileAccess=
		new EasyAccessSolitaryFile(informationFilePath,false);
	informationFileAccess.skipBOM();
	dictionaryKeysLanguage=informationFileAccess.readUTF8String();
	dictionaryArticlesLanguage=informationFileAccess.readUTF8String();
	dictionarySource=informationFileAccess.readUTF8String();
    }catch(FileNotFoundException e){
	dictionaryArticlesLanguage=dictionaryKeysLanguage=sourceFolder.getName();
	dictionarySource="Unknown source";
	System.err.println("Error, file \""+informationFilePath+
		"\" should exist. Dictionary keys language \""+
		dictionaryKeysLanguage+"\", dictionary articles language \""+
		dictionaryArticlesLanguage+"\" and dictionary source \""+
		dictionarySource+"\" would be written.");
    }
    int alphabetSize=Index.getAlphabetSize();
    String wordsFilePath;
    String wordsTranslationsFilePath;
    EasyAccessSolitaryFile wordsAccess;
    EasyAccessSolitaryFile wordsTranslationsAccess;
    StringBuffer sbWord;
    StringBuffer sbArticle;
    StringBuffer sbSourceWordOffset;
    String sSourceWordOffset;
    File wordsTranslationsFolder;
    for(j=0;j<alphabetSize;j++){
	letterFolderPath=sourceFolder.getPath()+File.separatorChar+j;
	letterFolder=new File(letterFolderPath);
	if(!letterFolder.exists()||letterFolder.isFile())
	    continue;
	wordsFolderPath=letterFolderPath+File.separatorChar+
		"words";
	wordsFolder=new File(wordsFolderPath);
	if(!wordsFolder.exists()||wordsFolder.isFile()){
	    System.err.println("Error, folder \""+wordsFolderPath+
		    "\" should exist.");
	    continue;
	}
	wordsTranslationsFolderPath=letterFolderPath+
		File.separatorChar+"tran";
	wordsTranslationsFolder=new File(
		wordsTranslationsFolderPath);
	if(!wordsTranslationsFolder.exists()||
		wordsTranslationsFolder.isFile()){
	    System.err.println("Error, folder \""+
		    wordsTranslationsFolderPath+
		    "\" should exist.");
	    continue;
	}
	for(i=0;i<alphabetSize;i++){
	    wordsFilePath=wordsFolderPath+File.separatorChar+i+
		    "_word.txt";
	    wordsTranslationsFilePath=wordsTranslationsFolderPath+
		    File.separatorChar+i+"_tran.txt";
	    try{
		wordsAccess=new EasyAccessSolitaryFile(wordsFilePath,
			false);
	    }catch(FileNotFoundException e){
		continue;
	    }
	    try{
		wordsTranslationsAccess=new EasyAccessSolitaryFile(
			wordsTranslationsFilePath,false);
	    }catch(FileNotFoundException e){
		System.err.println("Error, file \""+wordsTranslationsFilePath+
			"\" should exist.");
		continue;
	    }
	    wordsAccess.skipBOM();
	    while(true){
		sbWord=new StringBuffer();
		try{
		    while((nCharacter=wordsAccess.readUTF8Char())!=':')
			sbWord.append((char)nCharacter);
		}catch(EOFException e){
		    break;
		}
		sbSourceWordOffset=new StringBuffer();
		try{
		    while((nCharacter=wordsAccess.readUTF8Char())!=0x0D)
			if('0'<=nCharacter&&nCharacter<='9')
			    sbSourceWordOffset.append((char)nCharacter);
		}catch(EOFException e){
		}
		sSourceWordOffset=sbSourceWordOffset.toString();
		try{
		    wordsTranslationsAccess.seek(Integer.parseInt(
			    sSourceWordOffset));
		}catch(NumberFormatException e){
		    System.err.println("Error, \""+sSourceWordOffset+
			    "\" isn't like a number at \""+
			    wordsFilePath+"\".");
		    continue;
		}
		sbArticle=new StringBuffer();
		try{
		    while((nCharacter=
			    wordsTranslationsAccess.readUTF8Char())!=
			    SOURCE_FILE_ARTICLE_ENDING)
			sbArticle.append((char)nCharacter);
		}catch(EOFException e){
		}
		writeArticle(sbWord.toString(),sbArticle.toString());
	    }
	}
    }
    System.out.println("Source dictionary reading completed; "+Index.
	    getArticlesCount()+" articles found.");
    Index.writeAll(destinationFileAccess,dictionaryKeysLanguage,
	    dictionaryArticlesLanguage,dictionarySource);
    articlesFileAccess.seek(0);
    while(true)
	try{
	    destinationFileAccess.writeByte(
		    articlesFileAccess.readUnsignedByte());
	}catch(EOFException e){
	    break;
	}
    destinationFileAccess.close();
    System.out.println("Destination dictionary writing completed.");
}

private static void printArticle(
	EasyAccessFile destinationFileAccess,int articleOffset)
	throws IOException{
    char character;
    destinationFileAccess.seek(articleOffset);
    while((character=destinationFileAccess.readUTF8Char())!=
	    DESTINATION_FILE_ARTICLE_ENDING)
	System.out.print(character);
    System.out.println();
}

private static int printFreeArticles(
	EasyAccessFile destinationFileAccess,
	Vector<Integer> indexOffset) throws IOException{
    int printedArticlesCount=0;
    int articleOffset;
    int j;
    for(j=indexOffset.size()-1;j>=0;j--){
	destinationFileAccess.seek(indexOffset.get(j)+INT_LENGTH);
	articleOffset=destinationFileAccess.readInt();
	if(articleOffset==NULL_POINTER)
	    continue;
	printArticle(destinationFileAccess,articleOffset);
	printedArticlesCount++;
    }
    return printedArticlesCount;
}

private static void printHelp(){
    System.out.println(
	    "Usage:\n   java dictionarybuilder/Main SOURCE_FOLDER_PATH "+
	    "DESTINATION_FOLDER_PATH\n      Converts MobileLexicon dictionary "+
	    "from SOURCE_FOLDER_PATH to PieceOfDictionary lexical base at "+
	    "DESTINATION_FOLDER_PATH.\n   java dictionarybuilder/Main -f WORD "+
	    "DESTINATION_FOLDER_PATH\n      Finds WORD's article at "+
	    "PieceOfDictionary lexical base.");
}

private static void removeAnyFile(File file){
    if(file.isDirectory()){
	File[] subfiles=file.listFiles();
	int j;
	for(j=0;j<subfiles.length;j++)
	    removeAnyFile(subfiles[j]);
    }
    file.delete();
}

/**
 * Записывает статью во временный файл.
 * 
 * @param word
 *            Ключ к статье. Всегда проходит {@link String#trim()}.
 */
private static void writeArticle(String word,String article) throws
	IOException{
    //System.out.println(word+" = "+article);
    word=word.trim();
    Index[] index=Index.get(word,true);
    int j;
    for(j=0;j<index.length;j++)
	index[j].addArticleOffset((int)articlesFileAccess.getFilePointer());
    int wordLength=word.length();
    for(j=0;j<wordLength;j++)
	articlesFileAccess.writeUTF8Char(word.charAt(j));
    articlesFileAccess.writeUTF8Char(ARTICLE_KEY_AND_ARTICLE_TEXT_DIVIDER);
    int articleLength=article.length();
    for(j=0;j<articleLength;j++)
	articlesFileAccess.writeUTF8Char(article.charAt(j));
    /*
     * Дописываем символ, указующий на конец статьи.
     */
    articlesFileAccess.writeUTF8Char(DESTINATION_FILE_ARTICLE_ENDING);
}

}
