import java.io.*;

/**
 * 
 *
 * @author quasiyoke
 */
public abstract class EasyReadingFile{

protected long filePointer;

public abstract void close() throws IOException;

protected void finalize(){
    try{
	close();
    }catch(IOException e){
    }
}

public long getFilePointer(){
    return filePointer;
}

public abstract char readChar() throws IOException;

public abstract int readInt() throws IOException;

public abstract int readUnsignedByte() throws IOException;

/**
 * @return Прочитанный символ.
 * @exception EOFException Если поток достигает конца до прочтения всех
 * нужных байт.
 * @exception IOException Когда происходит ошибка ввода/вывода.
 */
public char readUTF8Char() throws IOException{
    int character;
    int nByte=readUnsignedByte();
    if((nByte&0x80)==0)
	/*
	 * Символы Unicode из диапазона '\u0000' —'\u007F', в котором лежит
	 * английский алфавит, записываются одним байтом, старший байт
	 * просто отбрасывается.
	 */
	return (char)nByte;
    else{
	if((nByte&0xE0)==0xC0)
	    /*
	     * Символы Unicode из диапазона '\u0080' —'\u07FF', в котором лежат
	     * наиболее распространенные символы национальных алфавитов,
	     * записываются двумя байтами следующим образом: символ Unicode с
	     * кодировкой 00000хххххуууууу записывается как 110ххххх10уууууу.
	     */
	    character=(nByte&0x1F)<<6;
	else
	    /*
	     * Остальные символы Unicode из диапазона '\u0800' —'\UFFFF'
	     * записываются тремя байтами по следующему правилу: символ Unicode
	     * с кодировкой xxxxyyyyyyzzzzzz записывается как
	     * 1110xxxx10yyyyyy10zzzzzz.
	     */
	    character=((nByte&0x0F)<<12)|(readUnsignedByte()&0x3F)<<6;
	return (char)(character|(readUnsignedByte()&0x3F));
    }
}

public String readUTF8String() throws IOException{
    StringBuffer stringBuffer=new StringBuffer();
    char character;
    while((character=readUTF8Char())<=0x000D);
    try{
	do
	    stringBuffer.append(character);
	while((character=readUTF8Char())>0x000D);
    }catch(EOFException e){
    }
    return stringBuffer.toString();
}

/**
 * @param terminatingCharacter Символ, которым должна завершиться строка.
 */
public String readUTF8String(char terminatingCharacter) throws IOException{
    StringBuffer stringBuffer=new StringBuffer();
    char character;
    while((character=readUTF8Char())!=terminatingCharacter)
	stringBuffer.append(character);
    return stringBuffer.toString();
}

/**
 * Выполняет перемещение указателя в указанное положение. Никогда не вызывает
 * {@link EOFException}.
 * 
 * @return Ссылку на себя.
 */
public abstract EasyReadingFile seek(long position) throws IOException;

/**
 * Выполняет перемещение указателя вперёд на учазанное число байт. Если
 * <code>bytesCount</code> меньше единицы, переход не осуществляется.
 * 
 * @return Ссылку на себя.
 */
public abstract EasyReadingFile skipBytes(int bytesCount) throws IOException;

/**
 * Пытается прочесть первые байты файла, и если они не составляют BOM-сигнатуру
 * 0xEF, 0xBB, 0xBF; или возникает {@link EOFException} (файл меньше трёх байт),
 * переходит в исходную позицию.
 * 
 * @return Ссылку на себя.
 */
public EasyReadingFile skipBOM() throws IOException{
    long position=getFilePointer();
    try{
	if(readUTF8Char()!=0xFEFF)
	    seek(position);
    }catch(EOFException e){
	seek(position);
    }
    return this;
}

}
