import org.xmlpull.v1.*;
import java.io.*;
import java.util.*;

/**
 * 
 *
 * @author quasiyoke
 */
public class Language implements Comparable{

private static Language currentLanguage;

private String fileName;

private Hashtable inscriptions;

private String isoName;

/**
 * Отсортированный массив всех PieceOfDictionary-языков интерфейса.
 */
private static Language[] languages;

private String name;

private static XmlPullParser parser;

private String translator;

private int[] version;



static{
    try{
        parser=PieceOfDictionary.xmlPullParserFactory.newPullParser();
    }catch(XmlPullParserException e){
        PieceOfDictionary.showErrorAndExit("Language parser initializing: "+e.
                toString());
    }
}

/**
 * @throws XmlPullParserException Если невозможно прочитать заголовок файла.
 */
private Language(String fileName) throws IOException,XmlPullParserException{
    InputStream languageInputStream=PieceOfDictionary.getInstance().getClass().
            getResourceAsStream(fileName);
    this.fileName=fileName;
    parser.setInput(languageInputStream,null);
    readLanguageHead();
}

public int compareTo(Object object){
    if(object instanceof Language){
        return isoName.compareTo(((Language)object).isoName);
    }else{
        return 1;
    }
}

private void flush() throws IOException{
    inscriptions=null;
}

public static String getCurrentLanguageInscription(String key){
    return currentLanguage.getInscription(key);
}

public static String getCurrentLanguageName(){
    return currentLanguage.name;
}

private String getInscription(String key){
    String inscription=(String)inscriptions.get(key);
    if(inscription==null){
        PieceOfDictionary.showError("Inscription \""+key+"\" not found.");
        return "null";
    }else{
        return inscription;
    }
}

public static Language[] getLanguages(){
    return languages;
}

public static String[] getLanguagesNames(){
    String[] languagesNames=new String[languages.length];
    int j;
    for(j=0;j<languages.length;j++){
        languagesNames[j]=languages[j].name;
    }
    return languagesNames;
}

private void read() throws IOException,XmlPullParserException{
    inscriptions=new Hashtable();
    parser.setInput(PieceOfDictionary.getInstance().getClass().
            getResourceAsStream(fileName),null);
    readLanguageHead();
    int attributesCount;
    int j;
    String key;
    String attributeName;
    String attributeValue;
    String inscription;
    while(parser.nextTag()==XmlPullParser.START_TAG){
        parser.require(XmlPullParser.START_TAG,"","inscription");
        attributesCount=parser.getAttributeCount();
        key=null;
        for(j=0;j<attributesCount;j++){
            attributeName=parser.getAttributeName(j);
            attributeValue=parser.getAttributeValue(j);
            if("key".equals(attributeName)){
                key=attributeValue;
                break;
            }else{
                PieceOfDictionary.showError("\""+fileName+
                        "\" inscriptions reading -- unknown attribute: "+
                        attributeName+"=\""+
                        attributeValue+"\"");
            }
        }
        inscription=parser.nextText();
        if(key==null||key.length()==0){
            PieceOfDictionary.showError("\""+fileName+
                    "\" inscriptions reading -- inscription \""+inscription+
                    "\" key not found.");
        }else{
            inscriptions.put(key,inscription);
        }
        parser.require(XmlPullParser.END_TAG,"","inscription");
    }
    parser.require(XmlPullParser.END_TAG,"","PieceOfDictionary_language");
}

private void readLanguageHead() throws
        XmlPullParserException,IOException{
    parser.nextTag();
    parser.require(XmlPullParser.START_TAG,"","PieceOfDictionary_language");
    int attributesCount=parser.getAttributeCount();
    String attributeName;
    String attributeValue;
    int j;
    for(j=0;j<attributesCount;j++){
        attributeName=parser.getAttributeName(j);
        attributeValue=parser.getAttributeValue(j);
        if("iso_name".equals(attributeName)){
            isoName=attributeValue;
        }else if("version".equals(attributeName)){
            version=PieceOfDictionary.getInstance().getVersion(attributeValue);
        }else{
            PieceOfDictionary.showError("\""+fileName+
                    "\" reading -- unknown attribute: "+attributeName+"=\""+
                    attributeValue+"\"");
        }
    }
    if(isoName==null){
        PieceOfDictionary.showError("\""+fileName+
                "\" reading -- not all necessary attributes found.");
    }
    if(version==null){
        PieceOfDictionary.showError("\""+fileName+
                "\" reading -- version information not found.");
    }else{
        if(PieceOfDictionary.compareVersions(version,PieceOfDictionary.
                getInstance().getVersion())==1){
            PieceOfDictionary.showError(
                    "\""+fileName+"\" reading -- file's version ("+
                    PieceOfDictionary.getVersion(version)+
                    ") is fresher than PieceOfDictionary's one ("+PieceOfDictionary.getVersion(PieceOfDictionary.getInstance().
                    getVersion())+
                    "). PieceOfDictionary will try to read this file -- errors are possible.");
        }
    }
    if(parser.nextTag()!=XmlPullParser.START_TAG){
        throw new XmlPullParserException(
                "Tags \"<translator>\" or \"<name>\" expected.");
    }
    if("translator".equals(parser.getName())){
        translator=parser.nextText();
        parser.require(XmlPullParser.END_TAG,"","translator");
        parser.nextTag();
    }
    parser.require(XmlPullParser.START_TAG,"","name");
    parser.next();
    name=parser.getText();
    parser.nextTag();
    parser.require(XmlPullParser.END_TAG,"","name");
}

/**
 * Меняет текущий язык программы. Вызывает методы инициализации зависящих от
 * локализации строк, дабы они сменили надписи.
 */
public static void setCurrent(String name){
    int j;
    if(currentLanguage==null||!currentLanguage.name.equals(name)){
        for(j=0;j<languages.length;j++){
            if(name.equals(languages[j].name)){
                if(currentLanguage!=null){
                    try{
                        currentLanguage.flush();
                    }catch(IOException e){
                        final String FLUSHING_ERROR_MESSAGE="Language \""+
                                currentLanguage.name+"\" flushing: ";
                        PieceOfDictionary.showError(FLUSHING_ERROR_MESSAGE+e.
                                toString());
                    }
                }
                currentLanguage=languages[j];
                final String INSCRIPTIONS_READING_ERROR_MESSAGE=
                        "Language inscriptions reading: ";
                try{
                    currentLanguage.read();
                }catch(IOException e){
                    PieceOfDictionary.showErrorAndExit(INSCRIPTIONS_READING_ERROR_MESSAGE+
                            e.toString());
                }catch(XmlPullParserException e){
                    PieceOfDictionary.showErrorAndExit(INSCRIPTIONS_READING_ERROR_MESSAGE+
                            e.toString());
                }
                MainCanvas.getInstance().initializeStrings();
                PieceOfDictionary.getInstance().initializeStrings();
                return;
            }
        }
        throw new NoSuchElementException("There is not such language: \""+name+
                "\"");
    }
}

public static void setLanguagesPaths(Vector languagesPaths){
    int j;
    int languagesPathsCount=languagesPaths.size();
    String languagePath;
    Vector languages=new Vector();
    for(j=0;j<languagesPathsCount;j++){
        languagePath=(String)languagesPaths.elementAt(j);
        try{
            languages.addElement(new Language(languagePath));
        }catch(IOException e){
            PieceOfDictionary.showError("Language head reading (file \""+
                    languagePath+"\"): "+e.toString());
        }catch(XmlPullParserException e){
            PieceOfDictionary.showError("Language head reading (file \""+
                    languagePath+"\"): "+e.toString());
        }
    }
    int languagesCount=languages.size();
    Language.languages=new Language[languagesCount];
    for(j=0;j<languagesCount;j++){
        Language.languages[j]=(Language)languages.elementAt(j);
    }
    PieceOfDictionary.sort(Language.languages);
}

public String toString(){
    String myCurrentLanguageName=Language.getCurrentLanguageInscription(
            "language_"+isoName);
    if(translator==null){
        return PieceOfDictionary.replace(getCurrentLanguageInscription(
                "language_information"),PieceOfDictionary.STRING_ESCAPE[0],
                myCurrentLanguageName);
    }else{
        return PieceOfDictionary.replace(PieceOfDictionary.replace(getCurrentLanguageInscription(
                "language_information_with_translator"),
                PieceOfDictionary.STRING_ESCAPE[0],myCurrentLanguageName),
                PieceOfDictionary.STRING_ESCAPE[1],translator);
    }
}

}
