import javax.microedition.lcdui.*;
import java.util.*;

/**
 * @author quasiyoke
 */
public final class MainCanvas extends Canvas implements CommandListener{

/**
 * Как много последовательностей клавиш следует хранить.
 */
private final static int KEY_CHAINS_HISTORY_SIZE=30;

private final static int KEY_CLEAR=-8;

/**
 * Как много миллисекунд считать долгим нажатием на экран.
 */
private final static int LONG_TAP_TIME=900;

/**
 * Шаг прокручивания сообщений.
 */
private final static int MESSAGES_SCROLLING_STEP=
        PieceOfDictionary.FONT_MESSAGES.getHeight()*1/5;

private String invitation;

private String messageNothingFound;

private LRUCache articlesCache;

private ArticlesScroller articlesScroller;

private boolean canScrollDown;

private Command commandClear;

private Command commandMenu;

/**
 * Индекс текущей последовательности нажатий в {@link MainCanvas#lastKeyChains}.
 */
private int currentKeyChainIndex;

public static boolean fullscreen;

/**
 * Отступ верхнего края сообщения-подсказки от верхнего края экрана.
 */
private int itemPositionY;

/**
 * Индекс элемента, с которого нужно начинать отрисовку. Чаще всего, это
 * высочайший рисуемый элемент.
 */
private int itemIndex=0;

private static MainCanvas instance;

/**
 * Текущая последовательность кодов клавиш.
 */
private Vector keyChain;

/**
 * Длина имеющей в данный момент значение части последовательности кодов клавиш.
 * При стирании последовательности с конца клавишей "Clear", например, коды
 * клавиш не удаляются реально, а лишь уменьшается значение этой переменной.
 */
private int keyChainLength;

/**
 * История последовательностей клавиш.
 */
private LimitedVector lastKeyChains;

private int mode;

private final static int MODE_HELP_MESSAGE_SHOWING=0;

private final static int MODE_NOTHING_FOUND=1;

private final static int MODE_ARTICLES=2;

/**
 * Насколько большую часть экрана считать несущественным перемещением при долгом
 * нажатии.
 */
private final static int SMALL_SCREEN_PART_FRACTION=(1<<
        PieceOfDictionary.POINT_POSITION)/20;

private LexicalBase.OccurrencesSet occurrences;

private boolean modeOrSizeChanged;

private Point screenSize=new Point();

private Point translation=new Point();

private MainCanvas(){
    keyChain=new Vector();
    lastKeyChains=new LimitedVector(KEY_CHAINS_HISTORY_SIZE);
    lastKeyChains.addElement(keyChain);
    currentKeyChainIndex=0;
    checkSize();
    setCommandListener(this);
    setFullScreenMode(fullscreen);
}

/**
 * Правильно добавляет клавишу к текущей их последовательности. Перерисовывает
 * экран если нужно. Заботится об истории последовательностей клавиш.
 */
private void addKey(char key){/* TODO Переписать ко всем собакам: уж больно задумчиво сделано. */
    if(!PieceOfDictionary.isSpace(key)){
        Character charKey=new Character(key);
        /*
         * Если осуществляется ввод в "историческую" последовательность клавиш,
         * или ввод новых символов в подстёртую последовательность, нужно
         * создавать новый элемент истории последовательностей нажатий путём
         * копирования.
         */
        if(currentKeyChainIndex>0||(keyChainLength<keyChain.size()&&!charKey.
                equals(keyChain.elementAt(keyChainLength)))){
            /*
             * Если осуществляется ввод в "историческую" последовательность
             * клавиш, а свежайшая последовательность пуста, использовать для
             * копирования нужно её.
             */
            if(currentKeyChainIndex>0){
                Vector lastKeyChain=(Vector)lastKeyChains.elementAt(0);
                if(lastKeyChain.isEmpty()){
                    keyChain=PieceOfDictionary.addAll(
                            keyChain,keyChainLength,lastKeyChain);
                }else{
                    keyChain=PieceOfDictionary.copyVector(keyChain,
                            keyChainLength);
                    lastKeyChains.addElement(keyChain);
                }
            }else{
                keyChain=PieceOfDictionary.copyVector(keyChain,keyChainLength);
                lastKeyChains.addElement(keyChain);
            }
            currentKeyChainIndex=0;
            keyChain.addElement(charKey);
        }else if(keyChainLength==keyChain.size()){
            keyChain.addElement(charKey);
        }
        keyChainLength++;
        processKeyChain();
    }
}

public void checkSize(){
    sizeChanged(getWidth(),getHeight());
}

/**
 * Производит "очистку" текущей последоваетльности клавиш, позволяя вводить новое
 * слово.
 */
private void clear(){
    /*
     * Создаём чистую последовательность клавиш только если свежайшая не такова.
     */
    if(((Vector)lastKeyChains.elementAt(0)).isEmpty()){
        if(currentKeyChainIndex<=1){
            new HistorySwitcher(false);
        }else{
            currentKeyChainIndex=0;
            keyChainLength=0;
            setMode(MODE_HELP_MESSAGE_SHOWING);
        }
    }else{
        keyChain=new Vector();
        lastKeyChains.addElement(keyChain);
        /*
         * Если находимся в конце истории, запускаем красивый-прекрасивый эффект
         * перелистывания.
         */
        if(currentKeyChainIndex==0){
            currentKeyChainIndex=1;
            new HistorySwitcher(false);
        }else{
            currentKeyChainIndex=0;
            keyChainLength=0;
            setMode(MODE_HELP_MESSAGE_SHOWING);
        }
    }
}

/**
 * Called when action should be handled
 */
public void commandAction(Command command,Displayable displayable){
    if(command==commandClear){
        clear();
    }else if(command==commandMenu){
        PieceOfDictionary.getInstance().showMenu();
    }
}

public static MainCanvas getInstance(){
    return instance==null?instance=new MainCanvas():instance;
}

public int getKeyChainLength(){
    return keyChainLength;
}

private String getKeyChainString(){
    StringBuffer sbString=new StringBuffer();
    int j;
    for(j=0;j<keyChainLength;j++){
        sbString.append(((Character)keyChain.elementAt(j)).charValue());
    }
    return sbString.toString();
}

/**
 * Инициализирует все зависимые от локализации поля. Вызывает {@link MainCanvas#showNecessaryCommands()}.
 */
public void initializeStrings(){
    removeCommand(commandClear);
    removeCommand(commandMenu);
    commandClear=new Command(Language.getCurrentLanguageInscription(
            "clear_button"),Command.CANCEL,1);
    commandMenu=
            new Command(Language.getCurrentLanguageInscription("menu_button"),
            Command.SCREEN,2);
    messageNothingFound=Language.getCurrentLanguageInscription(
            "nothing_found_message");
    invitation=Language.getCurrentLanguageInscription("invitation");
    processKeyChain();
    showNecessaryCommands();
}

/**
 * Called when a key is pressed.
 */
protected void keyPressed(int keyCode){
}

/**
 * Called when a key is released.
 */
protected void keyReleased(int keyCode){
    processKey(keyCode,false);
}

/**
 * Called when a key is repeated (held down).
 */
protected void keyRepeated(int keyCode){
    processKey(keyCode,true);
}

public void paint(Graphics g){
    g.translate(-g.getTranslateX(),-g.getTranslateY());
    g.setColor(PieceOfDictionary.COLOR_BACKGROUND);
    if(translation.x>0){
        g.fillRect(0,0,translation.x,screenSize.y);
    }else{
        g.fillRect(translation.x+screenSize.x,0,screenSize.x,screenSize.y);
    }
    g.translate(translation.x,translation.y);
    int j;
    Point position=
            new Point();
    switch(mode){
    case MODE_HELP_MESSAGE_SHOWING:
        if(modeOrSizeChanged){
            position.x=PieceOfDictionary.MARGIN_LEFT;
            position.y=screenSize.y;
            PieceOfDictionary.print(g,invitation,position,
                    PieceOfDictionary.FONT_MESSAGES,0,screenSize);
            final int HELP_MESSAGE_HEIGHT=position.y+
                    PieceOfDictionary.FONT_MESSAGES.getHeight()-screenSize.y;
            if(HELP_MESSAGE_HEIGHT>screenSize.y){
                new MessageScroller(
                        HELP_MESSAGE_HEIGHT,
                        MODE_HELP_MESSAGE_SHOWING);
            }else{
                itemPositionY=screenSize.y-HELP_MESSAGE_HEIGHT>>1;
            }
        }
        g.setColor(PieceOfDictionary.COLOR_BACKGROUND);
        g.fillRect(0,0,screenSize.x,itemPositionY);
        position.x=PieceOfDictionary.MARGIN_LEFT;
        position.y=itemPositionY;
        PieceOfDictionary.print(g,invitation,position,
                PieceOfDictionary.FONT_MESSAGES,PieceOfDictionary.COLOR_MESSAGES,
                screenSize);
        position.y+=PieceOfDictionary.FONT_MESSAGES.getHeight();
        g.setColor(PieceOfDictionary.COLOR_BACKGROUND);
        g.fillRect(0,position.y,screenSize.x,screenSize.y);
        break;
    case MODE_NOTHING_FOUND:
        String nothingFoundMessage=
                PieceOfDictionary.replace(messageNothingFound,
                PieceOfDictionary.STRING_ESCAPE[0],getKeyChainString());
        if(modeOrSizeChanged){
            position.x=PieceOfDictionary.MARGIN_LEFT;
            position.y=screenSize.y;
            PieceOfDictionary.print(g,nothingFoundMessage,position,
                    PieceOfDictionary.FONT_MESSAGES,0,screenSize);
            final int NOTHING_FOUND_MESSAGE_HEIGHT=position.y+
                    PieceOfDictionary.FONT_MESSAGES.getHeight()-screenSize.y;
            if(NOTHING_FOUND_MESSAGE_HEIGHT>screenSize.y){
                new MessageScroller(
                        NOTHING_FOUND_MESSAGE_HEIGHT,
                        MODE_NOTHING_FOUND);
            }else{
                itemPositionY=screenSize.y-NOTHING_FOUND_MESSAGE_HEIGHT>>1;
            }
        }
        g.setColor(PieceOfDictionary.COLOR_BACKGROUND);
        g.fillRect(0,0,screenSize.x,itemPositionY);
        position.x=PieceOfDictionary.MARGIN_LEFT;
        position.y=itemPositionY;
        PieceOfDictionary.print(g,nothingFoundMessage,position,
                PieceOfDictionary.FONT_MESSAGES,PieceOfDictionary.COLOR_MESSAGES,
                screenSize);
        position.y+=PieceOfDictionary.FONT_MESSAGES.getHeight();
        g.setColor(PieceOfDictionary.COLOR_BACKGROUND);
        g.fillRect(0,position.y,screenSize.x,screenSize.y);
        break;
    case MODE_ARTICLES:/* TODO Разобраться с прокруткой стилусом. */
        try{
            while(itemPositionY>0&&itemIndex>0){
                position.y=0;
                ((Dictionary.Head.Article)articlesCache.elementAt(itemIndex)).
                        print(g,position,
                        screenSize);
                itemPositionY-=position.y;
                itemIndex--;
            }
        }catch(IndexOutOfSomeBoundsException e){
        }
        position.y=itemPositionY;
        j=itemIndex;
        g.setColor(PieceOfDictionary.COLOR_BACKGROUND);
        g.fillRect(0,0,screenSize.x,itemPositionY);
        try{
            do{
                ((Dictionary.Head.Article)articlesCache.elementAt(j)).print(g,
                        position,screenSize);
                j++;
                if(articlesScroller==null&&position.y<0){
                    itemPositionY=position.y;
                    itemIndex=j;
                }
            }while(position.y<screenSize.y);
            canScrollDown=true;
        }catch(IndexOutOfSomeBoundsException e){
            canScrollDown=false;
            g.setColor(PieceOfDictionary.COLOR_BACKGROUND);
            g.fillRect(0,position.y,screenSize.x,screenSize.y);
        }
        if(articlesScroller==null&&repainter==null&&itemIndex==0&&itemPositionY>
                0){
            new ArticlesScroller(0);
        }
        break;
    }
    if(modeOrSizeChanged){
        modeOrSizeChanged=false;
    }
}

/**
 * Масштаб прокручивания суть отношение дистанции, на которую переместился экран,
 * к дистанции, покрытой пальцем.
 */
private final static int SCROLLING_SCALE_FRACTION=(2<<
        PieceOfDictionary.POINT_POSITION)/1;

private int previousItemPositionYFraction;

private int previousX;

private int previousY;

/**
 * Перерисовщик экрана, запускаемый, когда произошло касание экрана стилусом.
 * <code>null</code> если стилус не касается экрана сейчас.
 */
private Repainter repainter;

/**
 * Called when the pointer is dragged.
 */
protected void pointerDragged(int x,int y){
    int dX=x-previousX;
    int dY=y-previousY;
    itemPositionY=previousItemPositionYFraction+dY*SCROLLING_SCALE_FRACTION>>
            PieceOfDictionary.POINT_POSITION;
    int smallScreenPart=screenSize.getLeast()*SMALL_SCREEN_PART_FRACTION>>
            PieceOfDictionary.POINT_POSITION;
    if(longTapDetector!=null&&dX*dX+dY*dY>smallScreenPart*smallScreenPart){
        longTapDetector=null;
    }
}

/**
 * Called when the pointer is pressed.
 */
protected void pointerPressed(int x,int y){
    if(historySwitcher!=null){
        historySwitcher.stop();
    }
    if(articlesScroller!=null){
        articlesScroller.stop();
    }
    previousX=x;
    previousY=y;
    previousItemPositionYFraction=itemPositionY<<
            PieceOfDictionary.POINT_POSITION;
    repainter=new Repainter();
    repainter.start();
    new LongTapDetector();
}

/**
 * Часть ширины экрана, которую должен преодолеть стилус для того, чтобы
 * произошло перелистывание истории.
 */
private final static int HISTORY_SWITCHING_THRESHOLD_FRACTION=(1<<
        PieceOfDictionary.POINT_POSITION)/2;

/**
 * Called when the pointer is released.
 */
protected void pointerReleased(int x,int y){
    repainter.stop();
    int dX=x-previousX;
    int dY=y-previousY;
    if(PieceOfDictionary.abs(dX)>PieceOfDictionary.abs(dY)&&PieceOfDictionary.
            abs(dX)>screenSize.x*HISTORY_SWITCHING_THRESHOLD_FRACTION>>
            PieceOfDictionary.POINT_POSITION){
        if(mode!=MODE_ARTICLES){
            itemPositionY=previousItemPositionYFraction>>
                    PieceOfDictionary.POINT_POSITION;
        }
        if(dX<0&&currentKeyChainIndex==0){
            clear();
        }else{
            new HistorySwitcher(dX>0);
        }
    }else{
        if(mode!=MODE_ARTICLES){
            itemPositionY=previousItemPositionYFraction>>
                    PieceOfDictionary.POINT_POSITION;
            repaint();
        }else{
            itemPositionY=previousItemPositionYFraction+dY*
                    SCROLLING_SCALE_FRACTION>>
                    PieceOfDictionary.POINT_POSITION;
            if(itemPositionY>0){
                new ArticlesScroller(0);
            }else{
                repaint();
            }
        }
    }
}

/**
 * На насколько большую часть экрана должна осуществляться прокрутка при
 * нажатии клавиш вверх/вниз.
 */
private static int KEY_SCROLLING_SCREEN_FRACTION=(2<<
        PieceOfDictionary.POINT_POSITION)/3;

/**
 * Производит обработку нажатий клавиш. Перерисовывает экран если нужно.
 */
private void processKey(int key,boolean keyRepeated){
    if(historySwitcher!=null){
        return;
    }
    int gameAction=getGameAction(key);
    if(mode==MODE_ARTICLES&&(gameAction==UP||gameAction==DOWN)&&key<0){
        if(articlesScroller!=null){
            articlesScroller.stop();
        }
        if(gameAction==UP){
            itemPositionY+=screenSize.y*KEY_SCROLLING_SCREEN_FRACTION>>
                    PieceOfDictionary.POINT_POSITION;
        }else if(canScrollDown){
            itemPositionY-=screenSize.y*KEY_SCROLLING_SCREEN_FRACTION>>
                    PieceOfDictionary.POINT_POSITION;
        }
        repaint();
    }else if(key<0&&(gameAction==LEFT||gameAction==RIGHT)){
        if(gameAction==RIGHT&&currentKeyChainIndex==0){
            clear();
        }else{
            new HistorySwitcher(gameAction==LEFT);
        }
    }else if(keyRepeated){
        clear();
    }else if(KEY_NUM1<=key&&key<=KEY_NUM9){
        addKey((char)key);
    }else if(key==KEY_CLEAR){
        if(keyChainLength>0){
            if(--keyChainLength==0){
                clear();
            }else{
                if(currentKeyChainIndex>0){
                    keyChain=PieceOfDictionary.copyVector(keyChain,
                            keyChainLength);
                    lastKeyChains.addElement(keyChain);
                    currentKeyChainIndex=0;
                }
                processKeyChain();
            }
        }
    /*
     * Клавиша с Unicode-символом.
     */
    }else if(key>0){
        addKey((char)key);
    }
/*else
PieceOfDictionary.getDisplay().setCurrent(new Alert(Integer.toString(key)),this);*/

}

/**
 * Обрабатывает последовательность нажатых клавиш (вызывается при необходимости).
 * Если надо, перерисовывает экран.
 */
public void processKeyChain(){
    if(keyChainLength==0){
        setMode(MODE_HELP_MESSAGE_SHOWING);
    }else{
        occurrences=LexicalBase.getInstance().new OccurrencesSet(keyChain,
                keyChainLength);
        articlesCache.setList(occurrences);
        try{
            occurrences.elementAt(0);
            setMode(MODE_ARTICLES);
        }catch(Exception e){
            setMode(MODE_NOTHING_FOUND);
        }
    }
}

/**
 * Вводит холст в нужный режим. Если надо, делает перерисовку.
 */
private void setMode(int mode){
    switch(mode){
    case MODE_HELP_MESSAGE_SHOWING:
        break;
    case MODE_NOTHING_FOUND:
        break;
    case MODE_ARTICLES:
        itemPositionY=0;
        itemIndex=0;
        canScrollDown=true;
        break;
    }
    if(mode!=this.mode){
        this.mode=mode;
    }
    showNecessaryCommands();
    repaint();
}

private void showNecessaryCommands(){
    addCommand(commandMenu);
    switch(mode){
    case MODE_HELP_MESSAGE_SHOWING:
        removeCommand(commandClear);
        break;
    case MODE_NOTHING_FOUND:
    case MODE_ARTICLES:
        addCommand(commandClear);
        break;
    }
    modeOrSizeChanged=true;
}

protected void sizeChanged(int width,int height){
    screenSize.x=width;
    screenSize.y=height;
    articlesCache=new LRUCache((height>width?height:width)*3/
            PieceOfDictionary.FONT_NORMAL_HEIGHT,occurrences);
    modeOrSizeChanged=true;
}

private final static int FRAMES_PER_SECOND=16;

/**
 * При прокрутке за край цепочки статей к словам, используется для возврещения
 * к краю экрана.
 */
class ArticlesScroller extends Thread{

private final static int SCROLLING_TIME=400;

private final static int STEPS_COUNT=SCROLLING_TIME*FRAMES_PER_SECOND/1000;

private final int DESTINATION_ITEM_POSITION_Y;

private final int START_ITEM_POSITION_Y;

boolean isRunning=true;



{
    articlesScroller=this;
}

/**
 * Создаёт, тут же запуская, новый прокручиватель статей.
 * 
 * @param destinationItemPositionY Значение, которое {@link
 * MainCanvas#itemPositionY} должна принять через {@link
 * ArticlesScroller#SCROLLING_TIME} миллисекунд.
 */
private ArticlesScroller(int destinationItemPositionY){
    DESTINATION_ITEM_POSITION_Y=destinationItemPositionY;
    START_ITEM_POSITION_Y=itemPositionY;
    start();
}

public void run(){
    for(int step=1;step<=STEPS_COUNT;step++){
        try{
            sleep(1000/FRAMES_PER_SECOND);
        }catch(InterruptedException e){
        }
        if(!isRunning){
            return;
        }
        itemPositionY=(START_ITEM_POSITION_Y*(STEPS_COUNT-step)+
                DESTINATION_ITEM_POSITION_Y*step)/STEPS_COUNT;
        repaint();
    }
    if(articlesScroller==this){
        articlesScroller=null;
    }
}

private void stop(){
    isRunning=false;
    articlesScroller=null;
}

}

private HistorySwitcher historySwitcher;

/**
 * Выполняет переходы меж элементами истории.
 */
class HistorySwitcher extends Thread{

private final static int CANT_SWITCH_SCREEN_FRACTION=(1<<
        PieceOfDictionary.POINT_POSITION)/2;

private final Vector HISTORICAL_KEY_CHAIN;

private final int DESTINATION_KEY_CHAIN_INDEX;

private final int PAUSE=1000/FRAMES_PER_SECOND;

private final boolean SWITCH_BACK;

private final static int SWITCHING_TIME=400;

/**
 * @throws IndexOutOfSomeBoundsException Если невозможен такой переход по истории.
 */
private HistorySwitcher(boolean switchBack){
    if(switchBack){
        HISTORICAL_KEY_CHAIN=(Vector)lastKeyChains.elementAt(
                currentKeyChainIndex+1);
        if(HISTORICAL_KEY_CHAIN!=null){
            DESTINATION_KEY_CHAIN_INDEX=currentKeyChainIndex+1;
        }else{
            DESTINATION_KEY_CHAIN_INDEX=currentKeyChainIndex;
        }
    }else{
        HISTORICAL_KEY_CHAIN=(Vector)lastKeyChains.elementAt(
                currentKeyChainIndex-1);
        if(HISTORICAL_KEY_CHAIN!=null){
            DESTINATION_KEY_CHAIN_INDEX=currentKeyChainIndex-1;
        }else{
            DESTINATION_KEY_CHAIN_INDEX=currentKeyChainIndex;
        }
    }
    SWITCH_BACK=switchBack;
    start();
    historySwitcher=this;
}

public void run(){
    final int STEPS_COUNT=SWITCHING_TIME*FRAMES_PER_SECOND/1000;
    int j;
    if(HISTORICAL_KEY_CHAIN==null){
        if(SWITCH_BACK){
            for(j=0;j<STEPS_COUNT>>1;j++){
                try{
                    sleep(PAUSE);
                }catch(InterruptedException e){
                }
                if(historySwitcher==null){
                    return;
                }
                translation.x=(screenSize.x*CANT_SWITCH_SCREEN_FRACTION>>
                        PieceOfDictionary.POINT_POSITION)*j/(STEPS_COUNT>>1);
                repaint();
            }
            for(;j>=0;j--){
                try{
                    sleep(PAUSE);
                }catch(InterruptedException e){
                }
                if(historySwitcher==null){
                    return;
                }
                translation.x=(screenSize.x*CANT_SWITCH_SCREEN_FRACTION>>
                        PieceOfDictionary.POINT_POSITION)*j/(STEPS_COUNT>>1);
                repaint();
            }
        }else{
            for(j=0;j<STEPS_COUNT>>1;j++){
                try{
                    sleep(PAUSE);
                }catch(InterruptedException e){
                }
                if(historySwitcher==null){
                    return;
                }
                translation.x=-(screenSize.x*CANT_SWITCH_SCREEN_FRACTION>>
                        PieceOfDictionary.POINT_POSITION)*j/(STEPS_COUNT>>1);
                repaint();
            }
            for(;j>=0;j--){
                try{
                    sleep(PAUSE);
                }catch(InterruptedException e){
                }
                if(historySwitcher==null){
                    return;
                }
                translation.x=-(screenSize.x*CANT_SWITCH_SCREEN_FRACTION>>
                        PieceOfDictionary.POINT_POSITION)*j/(STEPS_COUNT>>1);
                repaint();
            }
        }
    }else{
        if(SWITCH_BACK){
            for(j=0;j<STEPS_COUNT>>1;j++){
                try{
                    sleep(PAUSE);
                }catch(InterruptedException e){
                }
                if(historySwitcher==null){
                    return;
                }
                translation.x=screenSize.x*j/(STEPS_COUNT>>1);
                repaint();
            }
            for(j=0;j<=STEPS_COUNT>>1;j++){
                try{
                    sleep(PAUSE);
                }catch(InterruptedException e){
                }
                if(historySwitcher==null){
                    return;
                }
                translation.x=screenSize.x*j/(STEPS_COUNT>>1)-screenSize.x;
                if(j==0){
                    keyChain=HISTORICAL_KEY_CHAIN;
                    keyChainLength=keyChain.size();
                    currentKeyChainIndex=DESTINATION_KEY_CHAIN_INDEX;
                    processKeyChain();
                }else{
                    repaint();
                }
            }
        }else{
            for(j=0;j<STEPS_COUNT>>1;j++){
                try{
                    sleep(PAUSE);
                }catch(InterruptedException e){
                }
                if(historySwitcher==null){
                    return;
                }
                translation.x=-screenSize.x*j/(STEPS_COUNT>>1);
                repaint();
            }
            for(j=0;j<=STEPS_COUNT>>1;j++){
                try{
                    sleep(PAUSE);
                }catch(InterruptedException e){
                }
                if(historySwitcher==null){
                    return;
                }
                translation.x=screenSize.x-screenSize.x*j/(STEPS_COUNT>>1);
                if(j==0){
                    keyChain=HISTORICAL_KEY_CHAIN;
                    keyChainLength=keyChain.size();
                    currentKeyChainIndex=DESTINATION_KEY_CHAIN_INDEX;
                    processKeyChain();
                }else{
                    repaint();
                }
            }
        }
    }
    historySwitcher=null;
}

private void stop(){
    historySwitcher=null;
    translation.x=0;
    if(HISTORICAL_KEY_CHAIN!=null){
        keyChain=HISTORICAL_KEY_CHAIN;
        keyChainLength=keyChain.size();
        currentKeyChainIndex=DESTINATION_KEY_CHAIN_INDEX;
        processKeyChain();
    }else{
        repaint();
    }
}

}

private LongTapDetector longTapDetector;

class LongTapDetector extends Thread{

private LongTapDetector(){
    longTapDetector=this;
    start();
}

public void run(){
    try{
        sleep(LONG_TAP_TIME);
    }catch(InterruptedException e){
    }
    if(longTapDetector==this){
        fullscreen=!fullscreen;
        setFullScreenMode(fullscreen);
        PieceOfDictionary.getInstance().deleteStorageAndSaveOptions();
    }
}

}

/**
 * Прокручивает сообщения (когда те выходят за пределы экрана).
 */
class MessageScroller extends Thread{

private final int MESSAGE_HEIGHT;

private final int MODE;

private final static int LONG_SLEEPING_TIME=3000;

private final static int NORMAL_SLEEPING_TIME=300;

/**
 * Создаёт новую прокручивалку сообщений, тут же запуская её. Прокручивалка
 * завершит свою работу как только режим холста ({@link MainCanvas#mode})
 * изменится с заданного в аргументах конструктора.
 */
private MessageScroller(int messageHeight,int mode){
    MESSAGE_HEIGHT=messageHeight;
    MODE=mode;
    start();
}

public void run(){
    itemPositionY=0;
    while(true){
        try{
            sleep(itemPositionY==0||itemPositionY+MESSAGE_HEIGHT<screenSize.y?LONG_SLEEPING_TIME
                    :NORMAL_SLEEPING_TIME);
        }catch(InterruptedException e){
        }
        if(mode!=MODE){
            return;
        }
        if(itemPositionY+MESSAGE_HEIGHT<screenSize.y){
            itemPositionY=0;
        }else{
            itemPositionY-=MESSAGES_SCROLLING_STEP;
        }
        repaint();
    }
}

}

/**
 * Поток, перерисовывающий экран через равные интервалы времени (при
 * прокручивании стилусом).
 */
class Repainter extends Thread{

private Repainter(){
}

public void run(){
    do{
        repaint();
        try{
            sleep(1000/FRAMES_PER_SECOND);
        }catch(InterruptedException e){
        }
    }while(repainter!=null);
}

private void stop(){
    repainter=null;
}

}

}